#include <sys/types.h>
#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <netinet/in.h>

#include "flowsaver.h"
#include "flow.h"
#include "stats.h"
#include "signal.h"
#include "flowfile.h"
#include "ip.h"
#include "types.h"
#include "config.h"

static struct flow_tqh* savequeue;
static char* flowsavedir;
static uint32 flows_saved;
static uint32 saveq_threshhold;

void flowsaver_init( char* fsdir, uint32 sqthold )
{
    saveq_threshhold = sqthold;
    flowsavedir = fsdir;
    savequeue = (struct flow_tqh*)calloc( 1, sizeof( struct flow_tqh ) );
    TAILQ_INIT( savequeue );
    flows_saved = 0;
    stats.cflows_sq = 0;
}

void flowsaver_addFlow( struct flow_t* newflow )
{
#ifdef DEBUG2
   printf( "flowsaver_addflow " );
   flow_print( newflow ); printf( "\n" );
#endif

   TAILQ_INSERT_TAIL( savequeue, newflow, tq_element );
   flows_saved++;
   stats.cflows_sq++;

   if( flows_saved >= saveq_threshhold ) {
       flowsaver_flush();

       /* save Q before the offsetpointer in the file will overflow */ 
   } else if( (TAILQ_FIRST( savequeue )->starttime + 0xffe0) < newflow->starttime ) {  
       flowsaver_flush();
   }
}

void flowsaver_flush( void )
{
#ifdef DEBUG
   struct flow_t *curflow;

   TAILQ_FOREACH(curflow, savequeue, tq_element )
   {
      flow_print( curflow );
   }
#endif 

   flowsaver_write( savequeue );
   flowsaver_init( flowsavedir, saveq_threshhold ); 
}

void* flowsaver_write( struct flow_tqh* queue )
{
  char tmpfilename[42];   /*.u */
  char finalfilename[42]; /*.s */
  struct flow_t *curflow, *_curflow;
  struct flowfileheader fheader;
  struct saveflow sflow;
  int fd;

  if( TAILQ_EMPTY( queue ) ) return 0;

  /* fheader is the header of the flowfile!!! (knoersack) */
  flowfileheader_init( &fheader );
  fheader.starttime = TAILQ_FIRST( queue )->starttime;
  
  if ( flowsavedir != NULL ) {
    sprintf( tmpfilename,   "%s/flow.%ld.u", flowsavedir, fheader.starttime );
    sprintf( finalfilename, "%s/flow.%ld.s", flowsavedir,  fheader.starttime );
  } else {
    sprintf( tmpfilename,   "flow.%ld.u", fheader.starttime );
    sprintf( finalfilename, "flow.%ld.s", fheader.starttime );
  }

  printf( "N:Dumping flows into:%s\n", tmpfilename );
  fflush(stdout);

  fd = open( tmpfilename, O_RDWR|O_CREAT, S_IRWXU|S_IRWXG );

  if( fd < 0 ) {
   perror( "while writing flow" );
   exit(1);
  }

  /* we need space for the header */
  write( fd, &fheader, sizeof( struct flowfileheader ));
	 
  /* Walk through the stack */
  curflow = TAILQ_FIRST( queue );

  while( curflow )
  {
      fheader.stoptime = curflow->starttime;
      fheader.bytes   += ( curflow->sdbytes + curflow->dsbytes );
      fheader.pkts    += ( curflow->sdpkts  + curflow->dspkts  );
      fheader.flows++;

      sflow.typenr      = curflow->typenr;
      sflow.startoffset = htons( curflow->starttime - fheader.starttime );
      sflow.endoffset   = htons( curflow->uptime    - fheader.starttime );
      sflow.sdbytes     = htonl( curflow->sdbytes );
      sflow.sdpkts      = htonl( curflow->sdpkts );
      sflow.dsbytes     = htonl( curflow->dsbytes );
      sflow.dspkts      = htonl( curflow->dspkts );
      sflow.sip         = curflow->sip ;
      sflow.dip         = curflow->dip ;

      switch( curflow->typenr ) {
          case TCPTYPE:
              sflow.tcp_sport = curflow->tcp_sport;
              sflow.tcp_dport = curflow->tcp_dport;
              break;
          case UDPTYPE:
              sflow.udp_sport = curflow->udp_sport;
              sflow.udp_dport = curflow->udp_dport;
              break;
          case ICMPTYPE:
              sflow.icmp_type = curflow->icmp_type;
              sflow.icmp_subcode = curflow->icmp_subcode;
      }
      
      write( fd, &sflow, sizeof( struct saveflow ) ); 

      _curflow = curflow;
      curflow  = TAILQ_NEXT( curflow, tq_element );

      free( _curflow );
  }

  /* change byteorder in header */
  fheader.starttime = htonl( fheader.starttime );
  fheader.stoptime  = htonl( fheader.stoptime );
  fheader.bytes     = htonq( fheader.bytes );
  fheader.flows     = htonl( fheader.flows );
  fheader.pkts      = htonl( fheader.pkts );

  /* Update the header */
  lseek( fd, 0, SEEK_SET );  
  write( fd, &fheader, sizeof( struct flowfileheader ) );

  close(fd);

  free( queue );

  rename(tmpfilename,finalfilename);

#ifdef DEBUG
  printf( "TCPflows Backup finished...\n" );
#endif

  return 0;
}
