#include <sys/types.h>

#include <netinet/in.h>

#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "ip.h"
#include "eth.h"
#include "iphandler.h"
#include "flowsaver.h"
#include "stats.h"


static struct flow_lh *bucket;
static struct flow_tqh timequeue;
static uint32 hashsize;
static uint32 timeq_threshhold;
static uint32 flow_timeout;


void iphandler_init( uint32 hsize, uint32 timeqth, uint32 flowto )
{
    hashsize = hsize;
    timeq_threshhold = timeqth;
    flow_timeout = flowto;
    bucket = (struct flow_lh*)calloc( hashsize, sizeof( struct flow_lh ) );
//    bzero( bucket, HASHSIZE * sizeof( struct flow_lh ) );
    TAILQ_INIT( &timequeue );
}

void iphandler_flush( void )
{
    struct flow_t* firstflow = NULL;

    while( ( firstflow = TAILQ_FIRST( &timequeue ) ) ) {
       iphandler_move( firstflow );
    }
}

static uint32 iphandler_hash( const struct ip_t *packet )
{
  uint32 place;

  place  = HASHFUNC( packet->sip, hashsize );
  place += HASHFUNC( packet->dip, hashsize );

  switch( packet->typenr ) {
      case UDPTYPE:  
          place += HASHFUNC( packet->udp_sport, hashsize );
          place += HASHFUNC( packet->udp_dport, hashsize );
          break;

      case TCPTYPE:
          place += HASHFUNC( packet->tcp_sport, hashsize );
          place += HASHFUNC( packet->tcp_dport, hashsize );
          break;

      case ICMPTYPE:
          place += HASHFUNC( packet->icmp_type, hashsize );
          place += HASHFUNC( packet->icmp_subcode, hashsize );
           break;
               
  }

  place = place % hashsize;

  return place;
}

static int iphandler_match( const struct ip_t* packet, const struct flow_t* flow )
{
#define CMPPAIRS( s1, d1, s2, d2 ) packet->s1 == flow->d1 && packet->s2 == flow->d2
  int retval = 0;

  /* first we check the IPs */
  if( CMPPAIRS(sip,sip,dip,dip) )
      retval =  1;
  else if( CMPPAIRS( sip, dip, dip, sip ) )
      retval = -1;
  else return 0;

  switch( packet->typenr ) {

      case TCPTYPE:
          if( CMPPAIRS( tcp_sport, tcp_sport, tcp_dport, tcp_dport ) && retval == 1 )
              break;
          else if( CMPPAIRS( tcp_sport, tcp_dport, tcp_dport, tcp_sport ) && retval == -1 )
              break;
          else { retval = 0; break; }

      case UDPTYPE:
          if( CMPPAIRS( udp_sport, udp_sport, udp_dport, udp_dport ) && retval == 1 )
              break;
          else if( CMPPAIRS( udp_sport, udp_dport, udp_dport, udp_sport ) && retval == -1 )
              break;
          else { retval = 0; break; }

      case ICMPTYPE:
          if( CMPPAIRS( icmp_type, icmp_type, icmp_subcode, icmp_subcode ) )
              break;
          else { retval = 0; break; }
  }

  return retval;
#undef CMPPAIRS
}

void iphandler_addPacket( char *buf, uint32 ts )
{
    uint16 coll = 0;
    uint32 offset;
    struct flow_t       *curflow;	/* Current Flowpointer */
    struct flow_t       *newflow;	/* temporary pointer   */
    const struct ip_t *packet;
    signed char found = 0;

    packet = (const struct ip_t *)buf; 

    offset = iphandler_hash( packet );

    /* traverse the list connected to the bucket and find the right flow */
    LIST_FOREACH( curflow, &bucket[offset], l_element ) {
        if( packet->typenr == curflow->typenr ) {
           found = iphandler_match( packet, curflow );

           if( found ) break; 

           if( ++coll > 5 ) {
              printf( "Coll:%lu:%u...", offset, coll ); 
              flow_print( curflow );
           }
        }
    }

    if( !curflow ) {	/* no one found */
        newflow = (struct flow_t*)malloc( sizeof( struct flow_t ) );
        if( newflow == NULL ) {
            printf( "E: %s:%d: malloc breaks", __FILE__, __LINE__ );
            kill( getpid(), SIGTERM );
        }
        flow_init( newflow ); /* knoersack *likes* constructors!!! */
        newflow->starttime = ts;
        newflow->typenr    = packet->typenr;
        newflow->sip       = packet->sip;
        newflow->dip       = packet->dip;

        switch( packet->typenr ) {
            case TCPTYPE:
                newflow->tcp_sport = packet->tcp_sport;
                newflow->tcp_dport = packet->tcp_dport;
                break;

            case UDPTYPE:
                newflow->udp_sport = packet->udp_sport;
                newflow->udp_dport = packet->udp_dport;
                break;

            case ICMPTYPE:
                newflow->icmp_type = packet->icmp_type;
                newflow->icmp_subcode = packet->icmp_subcode;
                break;
        }

        LIST_INSERT_HEAD( &bucket[offset], newflow, l_element  );
        TAILQ_INSERT_TAIL(     &timequeue, newflow, tq_element );

        curflow = newflow;
        found = 1;
 
        stats.cflows_tq++;

        if( stats.cflows_tq > timeq_threshhold ) {
            iphandler_move( TAILQ_FIRST( &timequeue ) );
        }
    }

    /* update statistics */
    stats_update_iptype( packet->typenr, ntohs( packet->len ) );

    if( found == 1 ) {
       curflow->sdbytes += ntohs( packet->len );
       curflow->sdpkts++;
       curflow->uptime = ts;
    } else {
       curflow->dsbytes += ntohs( packet->len );
       curflow->dspkts++;
       curflow->uptime = ts;
    }

#ifdef DEBUG
    flow_print( curflow ); printf( "\n" );
#endif
}

void iphandler_move( struct flow_t* curflow )
{
    LIST_REMOVE( curflow, l_element );
    TAILQ_REMOVE( &timequeue, curflow, tq_element );
    flowsaver_addFlow( curflow );
    stats.cflows_tq--;

    return;
}


/* housekeeping... */

void iphandler_check( time_t t )
{
   struct flow_t *curflow;

 #ifdef DEBUG
 printf( "begin of timedqueue check\n" );
 #endif

   while( ( curflow = TAILQ_FIRST( &timequeue )  ) ) {  
      if( t - curflow->starttime < (int)flow_timeout) break;
      iphandler_move( curflow );
   }

 #ifdef DEBUG
 printf( "end of timedqueue check\n" );
 #endif
}

void parse_etherpkt( char* buf, uint32 ts )
{
  register const struct eth_t *ep;
  register const vlan_t *vlp;
  uint16 etype;

#ifndef NO_PPPOE
  register const pppoe_t *pppoe;
  register const char *ppppayload;
  unsigned int ppp_type;
#endif

#ifdef DEBUG
  int i;
#endif

  ep = (const struct eth_t *)buf;
  etype=htons(ep->type);

  switch( etype ) {
    case 0x8100:            /* VLAN processing ... */
      vlp = (const vlan_t *)buf;
      if (htons(vlp->type) == 0x0800)
      {
           /*  printf("Vland:%d\n",ntohs(vlp->vlanid)); */
        iphandler_addPacket( (char *)( buf + sizeof( struct eth_t ) + 4 ), ts );
      }
      else
      {
          /* printf("Other VLAN Frametype:%0x\n",htons(vlp->type); */
      }
      break;

#ifndef NO_PPPOE
    case 0x8864: /* PPPOe */
      pppoe = (const pppoe_t *)((char *)buf + sizeof( struct eth_t ) );
      ppppayload = (const char *)(buf + sizeof(struct eth_t) + sizeof(pppoe_t) );

       /* 
          PPP RFC .. the  last bit of last octet of ppp type  must be set, at other ppp type 
          octets last bit mustnt be set. PPP Type can be 1,2 or 4 Byte long 
       */

      for( ppp_type=0;! (ppp_type & 1) ; ) { 
        ppp_type = ppp_type << 8;
        ppp_type +=  (u_char) ppppayload[0];  
        ppppayload++;
      }

      switch( ppp_type ) {

        case 0xc021:
#ifdef PRINT_OTHER
          printf("N: PPP LCP Packet.. Dropping\n");
#endif
        break;

        case 0x21:  /* IP  */
          iphandler_addPacket((char *) ppppayload, ts );
        break;

        default:
          printf("N:Unknown PPPOe Code:%0x Sessid:%0x Len:%d PPPType:%0x\n", 
             pppoe->code, ntohs(pppoe->sessionid), ntohs(pppoe->len), ppp_type );
          break;
      }
    break;
#endif

    case 0x0800:	/* Plain Ethernet */
      iphandler_addPacket( (char *)( buf + sizeof( struct eth_t ) ), ts );
    break;

    default:
   
#ifdef PRINT_OTHER
      printf("N: Other Frametype:%0x\n",etype);
#endif
    break;

  }
}
